package com.espirit.core.samples.dataaccessplugin.soap;

import de.espirit.common.base.Logging;
import de.espirit.common.tools.Strings;

import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.util.Arrays;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;


/**
 * Utility for accessing properties based on single or dotted expressions.
 */
public class PropertyAccessingUtil {

	private static final Class<PropertyAccessingUtil> LOGGER = PropertyAccessingUtil.class;


	/**
	 * Evaluates a dotted expression and returns the property's value as a string.
	 *
	 * @param object The contextual object of the expression.
	 * @param dottedExpression The dotted expression to evaluate.
	 * @return The evaluated string, or {@code null},
	 */
	@Nullable
	public static String evaluateDottedExpressionAsString(@NotNull final Object object, @Nullable final String dottedExpression) {
		try {
			final Object evaluation = evaluateDottedExpression(object, dottedExpression);
			if (evaluation != null) {
				return evaluation.toString();
			}
		} catch (final Exception e) {
			Logging.logDebug("Evaluation of '" + dottedExpression + "' failed: " + e.toString(), LOGGER);
		}
		return null;
	}


	/**
	 * Evaluates a dotted expression and returns the property's value, or {@code null}.
	 *
	 * @param object The contextual object of the expression.
	 * @param dottedExpression The dotted expression to evaluate.
	 * @return The evaluated property, or {@code null},
	 * @throws IllegalStateException if the given expression could not be processed.
	 */
	@Nullable
	public static Object evaluateDottedExpression(@NotNull final Object object, @Nullable String dottedExpression) throws IllegalStateException {
		if (dottedExpression != null) {
			dottedExpression = dottedExpression.trim();
			if (!dottedExpression.isEmpty()) {
				final List<String> expressions = Arrays.asList(dottedExpression.split("\\."));
				return evaluateDottedExpression(object, expressions);
			}
		}
		return null;
	}


	/**
	 * Evaluates a list of chained expressions and returns the property's value, or {@code null}.
	 *
	 * @param object The contextual object of the expression.
	 * @param expressions The chained expressions to evaluate.
	 * @return The evaluated property, or {@code null},
	 * @throws IllegalStateException if the given expression could not be processed.
	 */
	@Nullable
	public static Object evaluateDottedExpression(@NotNull final Object object, @NotNull final List<String> expressions) throws IllegalStateException {
		Object tmp = object;
		for (int i = 0; i < expressions.size(); i++) {
			try {
				if (tmp instanceof PropertyAccessor) {
					tmp = evaluateMember((PropertyAccessor) tmp, expressions.get(i));
				} else {
					final String rest = Strings.implode(expressions.subList(i, expressions.size()), ".");
					throw new IllegalStateException("Failed to evaluate expression '" + rest + "' on '" + tmp + "'!");
				}
			} catch (final IllegalArgumentException e) {
				final String rest = Strings.implode(expressions.subList(i, expressions.size()), ".");
				throw new IllegalStateException("Failed to evaluate expression '" + rest + "' on '" + tmp + "'!", e);
			}
		}
		return tmp;
	}


	/**
	 * Evaluates a chained expression and sets the evaluated property's value.
	 *
	 * @param object The contextual object of the expression.
	 * @param expressions The chained expressions pointing to the property to modify.
	 * @param value The new property's value.
	 * @throws IllegalStateException if the given expression could not be processed.
	 */
	public static void evaluateDottedExpressionAndSetValue(@NotNull final Object object, @NotNull final List<String> expressions, @Nullable final Object value) throws IllegalStateException {
		if (expressions.isEmpty()) {
			return;
		}
		Object tmp = object;
		for (int i = 0; i < expressions.size() - 1; i++) {
			try {
				if (tmp instanceof PropertyAccessor) {
					tmp = evaluateMember((PropertyAccessor) tmp, expressions.get(i));
				} else {
					final String rest = Strings.implode(expressions.subList(i, expressions.size()), ".");
					throw new IllegalStateException("Failed to evaluate expression '" + rest + "' on '" + tmp + "'!");
				}
			} catch (final Exception e) {
				final String rest = Strings.implode(expressions.subList(i, expressions.size()), ".");
				throw new IllegalStateException("Failed to evaluate expression '" + rest + "' on '" + tmp + "'!", e);
			}
		}
		if (tmp instanceof PropertyAccessor) {
			final String name = expressions.get(expressions.size() - 1);
			((PropertyAccessor) tmp).setProperty(name, value);
		}
	}


	/**
	 * Evaluates a member of the specified property-holding object.
	 * This member might be a property (expression: "property"), or a property's list-value's entry (expression: "list[0]").
	 *
	 * @param accessor The accessor object to evaluate a member for.
	 * @param expression The expression targeting the member.
	 * @return The evaluated value, or {@code null}.
	 * @throws IllegalArgumentException If the expression is invalid.
	 */
	@Nullable
	public static Object evaluateMember(@NotNull final PropertyAccessor accessor, @NotNull final String expression) throws IllegalArgumentException {
		final Pattern pattern = Pattern.compile("([\\w\\-_]+)(\\[(\\d+)\\])?");
 		final Matcher matcher = pattern.matcher(expression);
		if (matcher.find()) {
			final String property = matcher.group(1);
			Object value = accessor.getProperty(property);
			final String index = matcher.group(3);
			if (index != null) {
				if (value != null && value.getClass().isArray()) {
					value = Arrays.asList((Object[]) value);
				}
				if (value instanceof List<?>) {
					return ((List<?>) value).get(Integer.parseInt(index));
				}
				return null;
			}
			return value;
		}
		throw new IllegalArgumentException("Invalid expression '" + expression + "'!");
	}
}
