package com.espirit.core.samples.dataaccessplugin.soap;

import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;


/**
 * Object handled by the {@link SOAPDataAccessPlugin}.
 */
public class SOAPObject {

	private final String _id;

	private ServiceClient _service;
	private PropertyAccessor _object;
	private boolean _null;


	public SOAPObject(@NotNull final String id) {
		_id = id;
	}


	/**
	 * Returns the unique identifier of this SOAP object.
	 *
	 * @return The identifier.
	 */
	@NotNull
	public String getId() {
		return _id;
	}


	/**
	 * Sets the service client instance this object might request it's underlying SOAP object with.
	 *
	 * @param service The service client to use.
	 */
	public void setService(@NotNull final ServiceClient service) {
		_service = service;
	}


	/**
	 * Sets the underlying SOAP object this object is associated with.
	 *
	 * @param object The underlying SOAP object of this instance.
	 */
	public void setObject(@NotNull final PropertyAccessor object) {
		_null = false;
		_object = object;
	}


	/**
	 * Returns the underlying SOAP object this instance is associated with.
	 * This method will request the object from the configured service client if necessary.
	 *
	 * @return The SOAP object this instance is associated with.
	 */
	@NotNull
	public PropertyAccessor getObject() {
		if (_object == null) {
			if (_service == null) {
				throw new IllegalStateException("Service not available!");
			}
			if (!_null) {
				_object = _service.getAccessorById(_id);
				_null = _object == null;
			}
			if (_object == null) {
				throw new IllegalStateException("Object with ID '" + _id + "' not found!");
			}
		}
		return _object;
	}


	/**
	 * Evaluates the snippet header of this object.
	 *
	 * @return The snippet header.
	 */
	@NotNull
	public String getSnippetHeader() {
		final ServiceConfiguration configuration = _service.getConfiguration();
		return evaluateString(configuration.operationById_evaluateHeader, _id);
	}


	/**
	 * Evaluates the snippet extract of this object.
	 *
	 * @return The snippet extract, or {@code null}.
	 */
	@Nullable
	public String getSnippetExtract() {
		final ServiceConfiguration configuration = _service.getConfiguration();
		return evaluateString(configuration.operationById_evaluateExtract, null);
	}


	/**
	 * Evaluates the snippet image url of this object.
	 *
	 * @return The snippet image url, or {@code null}.
	 */
	@Nullable
	public String getSnippetImage() {
		final ServiceConfiguration configuration = _service.getConfiguration();
		return evaluateString(configuration.operationById_evaluateImage, null);
	}


	/**
	 * Evaluates the given expression in the context of the underlying SOAP object.
	 *
	 * @param expression The expression to evaluate on the underlying SOAP object.
	 * @param fallback The fallback to use if the evaluation fails.
	 * @return The evaluated string, or the specified fallback.
	 */
	private String evaluateString(@Nullable final String expression, @Nullable final String fallback) {
		if (expression != null) {
			try {
				final PropertyAccessor object = getObject();
				final String evaluation = PropertyAccessingUtil.evaluateDottedExpressionAsString(object, expression);
				if (evaluation != null) {
					return evaluation;
				}
			} catch (final Exception ignored) {
			}
		}
		return fallback;
	}
}
