package com.espirit.core.samples.dataaccessplugin.soap;

import de.espirit.common.base.Logging;
import de.espirit.firstspirit.client.plugin.report.ParameterMap;

import org.apache.cxf.endpoint.Client;
import org.apache.cxf.endpoint.Endpoint;
import org.apache.cxf.jaxws.endpoint.dynamic.JaxWsDynamicClientFactory;
import org.apache.cxf.service.model.BindingInfo;
import org.apache.cxf.service.model.BindingOperationInfo;
import org.apache.cxf.service.model.ServiceInfo;
import org.jetbrains.annotations.ApiStatus;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import javax.tools.ToolProvider;
import javax.xml.namespace.QName;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.List;


/**
 * Client connection to a configured SOAP web service.
 */
public class ServiceClient {

	private static final Class<ServiceClient> LOGGER = ServiceClient.class;

	private final ServiceConfiguration _configuration;
	private final Client _client;
	private final SimpleLRUCache<String, SOAPObject> _objectCache;


	public ServiceClient(@NotNull final ServiceConfiguration configuration) {
		if (ToolProvider.getSystemJavaCompiler() == null) {
			throw new IllegalStateException("This SOAP service integration requires a JDK, it does not work with a JRE.");
		}
		_configuration = configuration;
		_client = JaxWsDynamicClientFactory.newInstance().createClient(_configuration.wsdlUrl);
		_objectCache = new SimpleLRUCache<>(1024);
	}


	/**
	 * Returns the service configuration this client is operating on.
	 *
	 * @return The service configuration.
	 */
	@NotNull
	ServiceConfiguration getConfiguration() {
		return _configuration;
	}


	/**
	 * Requests the {@link SOAPObject data object} associated with the specified ID.
	 *
	 * @param id The ID of the object to return.
	 * @return The object, or {@code null}.
	 */
	@Nullable
	public SOAPObject getObjectById(@NotNull final String id) {
		final SOAPObject cached = _objectCache.get(id);
		if (cached != null) {
			return cached;
		}
		final PropertyAccessor result = getAccessorById(id);
		if (result != null) {
			final SOAPObject object = new SOAPObject(id);
			object.setService(this);
			object.setObject(result);
			_objectCache.put(id, object);
			return object;
		}
		return null;
	}


	/**
	 * Provides a convenience object to access the concrete properties of SOAP serviced raw object
	 * associated with the specified ID.
	 *
	 * @param id The ID of the raw object to be accessed.
	 * @return An accessor object, or {@code null}, if no associated raw object was found.
	 */
	@Nullable
	PropertyAccessor getAccessorById(@NotNull final String id) {
		final BindingInfo binding = getBinding();
		final QName operationName = new QName(binding.getService().getTargetNamespace(), _configuration.operationById);
		final BindingOperationInfo operation = binding.getOperation(operationName);
		if (operation != null) {
			final ServiceOperation performable = new ServiceOperation(_client, operation);
			String expression = _configuration.operationById_evaluateSetId;
			if (expression != null) {
				expression = expression.trim();
				performable.evaluateSet(expression, id);
				return performable.perform();
			}
		}
		return null;
	}


	/**
	 * Requests a list of {@link SOAPObject data objects}s from the SOAP service. If given, parameters
	 * will be applied to filter the resulting list.
	 *
	 * @param parameters The parameters to be used as filter, or {@code null} for no filtering.
	 * @return The resulting list of {@link SOAPObject data object}s.
	 */
	@NotNull
	public List<SOAPObject> getObjects(@Nullable final ParameterMap parameters) {
		final BindingInfo binding = getBinding();
		final QName operationName = new QName(binding.getService().getTargetNamespace(), _configuration.operationList);
		final BindingOperationInfo operation = binding.getOperation(operationName);
		if (operation != null) {
			final ServiceOperation performable = new ServiceOperation(_client, operation);
			if (!initParameter(performable, parameters)) {
				return Collections.emptyList();
			}
			final PropertyAccessor response = performable.perform();
			if (response != null) {
				try {
					Object evaluation = PropertyAccessingUtil.evaluateDottedExpression(response, _configuration.operationList_evaluateEntry);
					if (evaluation != null) {
						if (evaluation.getClass().isArray()) {
							evaluation = Arrays.asList((Object[]) evaluation);
						}
						if (evaluation instanceof Collection) {
							final List<SOAPObject> list = new ArrayList<>();
							for (final Object entry : ((Collection<?>) evaluation)) {
								final String id = PropertyAccessingUtil.evaluateDottedExpressionAsString(entry, _configuration.operationList_evaluateEntryId);
								if (id != null) {
									final SOAPObject object = new SOAPObject(id);
									object.setService(this);
									list.add(object);
								}
							}
							return list;
						}
					}
				} catch (final IllegalStateException e) {
					Logging.logWarning("Response evaluation failed for operation '" + _configuration.operationList + "'!", e, LOGGER);
				}
			}
		}
		return Collections.emptyList();
	}


	/**
	 * Initializes the parameter configured for the list operation.
	 *
	 * @param performable The operation to initialize the parameter from.
	 * @param parameters The parameter values to use.
	 * @return Whether or not the parameter values has been applied successfully.
	 */
	private boolean initParameter(@NotNull final ServiceOperation performable, @Nullable final ParameterMap parameters) {
		if (parameters != null) {
			for (final ServiceConfiguration.ParameterConfiguration entry : _configuration.operationList_parameter) {
				final String name = entry.parameter.getName();
				final Object value = parameters.get(entry.parameter);
				String expression = entry.evaluateSetValue;
				if (expression != null) {
					expression = expression.trim();
					if (!expression.isEmpty()) {
						try {
							performable.evaluateSet(expression, value);
						} catch (final Exception e) {
							Logging.logWarning("Failed to apply parameter '" + name + "' of operation '" + _configuration.operationList + "'!", e, LOGGER);
							return false;
						}
					}
				}
			}
		}
		return true;
	}


	/**
	 * Return the associated client instance.
	 *
	 * @return The client instance.
	 */
	@NotNull
	Client getClient() {
		return _client;
	}


	/**
	 * Returns the configured service binding of the SOAP web service.
	 *
	 * @return The configured service binding.
	 */
	@NotNull
	BindingInfo getBinding() {
		final Endpoint endpoint = _client.getEndpoint();
		final String namespaceURI = _client.getEndpoint().getService().getName().getNamespaceURI();
		final QName bindingName = new QName(namespaceURI, _configuration.binding);
		for (final ServiceInfo serviceInfo : endpoint.getService().getServiceInfos()) {
			final BindingInfo binding = serviceInfo.getBinding(bindingName);
			if (binding != null) {
				return binding;
			}
		}
		throw new IllegalStateException("Binding not found!");
	}
}
