package com.espirit.core.samples.dataaccessplugin.soap;

import org.apache.ws.commons.schema.XmlSchemaComplexType;
import org.apache.ws.commons.schema.XmlSchemaElement;
import org.apache.ws.commons.schema.XmlSchemaParticle;
import org.apache.ws.commons.schema.XmlSchemaSequence;
import org.apache.ws.commons.schema.XmlSchemaSequenceMember;
import org.apache.ws.commons.schema.XmlSchemaType;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.beans.IntrospectionException;
import java.beans.PropertyDescriptor;
import java.lang.reflect.InvocationTargetException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Locale;


/**
 * Object providing access to the properties of dynamically created
 * SOAP-based XML structures.
 */
public class ServiceObjectProxy implements PropertyAccessor {

	private final Object _peer;
	private final XmlSchemaComplexType _schemaType;


	public ServiceObjectProxy(@NotNull final Object peer, @NotNull final XmlSchemaComplexType schemaType) {
		_peer = peer;
		_schemaType = schemaType;
	}


	/**
	 * Wrapps the specified object in a {@link ServiceObjectProxy} if necessary,
	 * or returns the object itself if not.
	 *
	 * @param peer The object to create a proxy for.
	 * @param schemaType The schema type describing the specified object.
	 * @return The object or a proxy.
	 */
	public static Object createProxyIfNecessary(final Object peer, @NotNull final XmlSchemaType schemaType) {
		if (peer != null) {
			if (peer instanceof Collection) {
				final List<Object> list = new ArrayList<>();
				for (final Object entry : (Collection<?>) peer) {
					list.add(createProxyIfNecessary(entry, schemaType));
				}
				return list;
			} else if (peer.getClass().isArray()) {
				final List<Object> list = new ArrayList<>();
				for (final Object entry : (Object[]) peer) {
					list.add(createProxyIfNecessary(entry, schemaType));
				}
				return list;
			} else if (schemaType instanceof XmlSchemaComplexType) {
				return new ServiceObjectProxy(peer, (XmlSchemaComplexType) schemaType);
			}
		}
		return peer;
	}


	@Override
	@Nullable
	public Object getProperty(@NotNull final String name) throws IllegalStateException {
		final XmlSchemaParticle particle = _schemaType.getParticle();
		if (particle instanceof XmlSchemaSequence) {
			for (final XmlSchemaSequenceMember member : ((XmlSchemaSequence) particle).getItems()) {
				if (member instanceof XmlSchemaElement) {
					final String memberName = ((XmlSchemaElement) member).getName();
					if (name.equals(((XmlSchemaElement) member).getName())) {
						final XmlSchemaType memberSchemaType = ((XmlSchemaElement) member).getSchemaType();
						try {
							final PropertyDescriptor memberDescriptor = new PropertyDescriptor(memberName, _peer.getClass());
							final Object memberValue = memberDescriptor.getReadMethod().invoke(_peer);
							return createProxyIfNecessary(memberValue, memberSchemaType);
						} catch (final InvocationTargetException | IntrospectionException | IllegalAccessException | IllegalArgumentException e) {
							throw new IllegalStateException("Accessing member '" + memberName + "' failed!", e);
						}
					}
				}
			}
		}
		return null;
	}


	@Override
	public void setProperty(@NotNull final String name, @Nullable Object value) throws IllegalStateException {
		final XmlSchemaParticle particle = _schemaType.getParticle();
		if (particle instanceof XmlSchemaSequence) {
			for (final XmlSchemaSequenceMember member : ((XmlSchemaSequence) particle).getItems()) {
				if (member instanceof XmlSchemaElement) {
					final String memberName = ((XmlSchemaElement) member).getName();
					if (name.equals(((XmlSchemaElement) member).getName())) {
						try {
							final PropertyDescriptor memberDescriptor = new PropertyDescriptor(memberName, _peer.getClass());
							final Class<?> expected = memberDescriptor.getReadMethod().getReturnType();
							if (expected.isEnum() && value instanceof String) {
								//noinspection rawtypes
								value = Enum.valueOf((Class) expected, ((String) value).toUpperCase(Locale.ENGLISH));
							}
							memberDescriptor.getWriteMethod().invoke(_peer, value);
						} catch (final InvocationTargetException | IntrospectionException | IllegalAccessException | IllegalArgumentException e) {
							throw new IllegalStateException("Accessing member '" + memberName + "' failed!", e);
						}
					}
				}
			}
		}
	}


	/**
	 * Returns the dynamically create object this proxy is operating on.
	 *
	 * @return The proxy's peer.
	 */
	@NotNull
	public Object getPeer() {
		return _peer;
	}
}
