package com.espirit.core.samples.dataaccessplugin.soap;

import de.espirit.common.base.Logging;

import org.apache.cxf.endpoint.Client;
import org.apache.cxf.service.model.BindingMessageInfo;
import org.apache.cxf.service.model.BindingOperationInfo;
import org.apache.cxf.service.model.MessagePartInfo;
import org.apache.ws.commons.schema.XmlSchemaAnnotated;
import org.apache.ws.commons.schema.XmlSchemaComplexType;
import org.apache.ws.commons.schema.XmlSchemaElement;
import org.apache.ws.commons.schema.XmlSchemaParticle;
import org.apache.ws.commons.schema.XmlSchemaSequence;
import org.apache.ws.commons.schema.XmlSchemaSequenceMember;
import org.apache.ws.commons.schema.XmlSchemaType;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.beans.IntrospectionException;
import java.beans.PropertyDescriptor;
import java.lang.reflect.GenericArrayType;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.ParameterizedType;
import java.lang.reflect.Type;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;


/**
 * Handler for executing an operation of a SOAP web service.
 */
public class ServiceOperation {

	private static final Class<ServiceOperation> LOGGER = ServiceOperation.class;

	private final Client _client;
	private final BindingOperationInfo _operation;
	private final Map<String, Object> _message;


	public ServiceOperation(@NotNull final Client client, @NotNull final BindingOperationInfo operation) {
		_client = client;
		_operation = operation;
		_message = new HashMap<>();
	}


	/**
	 * Evaluates a dotted expression and sets the evaluated request property's value.
	 *
	 * @param expression The dotted expression pointing to the request property to modify.
	 * @param value The new property's value.
	 */
	public void evaluateSet(@NotNull String expression, @Nullable final Object value) {
		expression = expression.trim();
		if (!expression.isEmpty()) {
			final List<String> expressionList = Arrays.asList(expression.split("\\."));
			// The first part of the dotted expression specified the message part itself
			final String partName = expressionList.get(0);
			if (expressionList.size() == 1) {
				// If there are no more expression the caller has specified the message part itself
				_message.put(partName, value);
			} else {
				// The rest of the expression points to the inner structure of the message part
				Object part = _message.get(partName);
				final BindingMessageInfo inputInfo = _operation.getInput();
				if (part == null) {
					// The related message part has not yet been created
					//  => Search for the message part description, and create one
					for (final MessagePartInfo partInfo : inputInfo.getMessageParts()) {
						if (partName.equals(partInfo.getName().getLocalPart())) {
							part = createMessagePart(partInfo);
							_message.put(partName, part);
							break;
						}
					}
				}
				if (part != null) {
					final List<String> expressionRest = expressionList.subList(1, expressionList.size());
					try {
						PropertyAccessingUtil.evaluateDottedExpressionAndSetValue(part, expressionRest, value);
					} catch (final Exception e) {
						throw new IllegalStateException("Failed to modify '" + expression + "'!", e);
					}
				} else {
					final String messageName = inputInfo.getMessageInfo().getName().getLocalPart();
					throw new IllegalStateException("Part '" + partName + "' of message '" + messageName + "' is no complex object to evaluate expressions on!");
				}
			}
		} else {
			throw new IllegalArgumentException("Expression must not be empty!");
		}
	}


	/**
	 * Executes the configuration using the previously configured request parameters.
	 *
	 * @return The response of the SOAP web service.
	 */
	@Nullable
	public PropertyAccessor perform() {
		final String operationName = _operation.getName().getLocalPart();
		try {
			final BindingMessageInfo input = _operation.getInput();
			// Collect all message parts and request parameters
			final List<Object> parameters = new ArrayList<>();
			for (final MessagePartInfo partInfo : input.getMessageParts()) {
				final String partName = partInfo.getName().getLocalPart();
				Object part = _message.get(partName);
				if (part == null) {
					part = createMessagePart(partInfo);
				}
				if (part instanceof ServiceObjectProxy) {
					part = ((ServiceObjectProxy) part).getPeer();
				}
				parameters.add(part);
			}
			final Object[] response;
			if (_operation.isUnwrapped()) {
				response = _client.invoke(_operation.getUnwrappedOperation(), parameters.toArray());
				final BindingMessageInfo output = _operation.getOutput();
				final ServiceObjectMap responseObject = new ServiceObjectMap();
				for (final MessagePartInfo partInfo : output.getMessageParts()) {
					final String partName = partInfo.getName().getLocalPart();
					final XmlSchemaElement partSchema = (XmlSchemaElement) partInfo.getXmlSchema();
					final XmlSchemaType schemaType = partSchema.getSchemaType();
					final Object partValue = ServiceObjectProxy.createProxyIfNecessary(response[0], schemaType);
					responseObject.setProperty(partName, partValue);
				}
				return responseObject;
			} else {
				response = _client.invokeWrapped(operationName, parameters.toArray());
				final BindingMessageInfo output = _operation.getOutput();
				final ServiceObjectMap responseObject = new ServiceObjectMap();
				for (final MessagePartInfo partInfo : output.getMessageParts()) {
					// In case of a wrapped operations, the response contain the result object's content,
					// instead of the result object itself.
					//  => So we need to create the result object first,
					//     and fill it with the response's objects.
					final String partName = partInfo.getName().getLocalPart();
					final XmlSchemaElement partSchema = (XmlSchemaElement) partInfo.getXmlSchema();
					final XmlSchemaType schemaType = partSchema.getSchemaType();
					if (schemaType instanceof XmlSchemaComplexType) {
						try {
							final XmlSchemaParticle particle = ((XmlSchemaComplexType) schemaType).getParticle();
							if (particle instanceof XmlSchemaSequence) {
								final List<XmlSchemaElement> children = new ArrayList<>();
								for (final XmlSchemaSequenceMember member : ((XmlSchemaSequence) particle).getItems()) {
									if (member instanceof XmlSchemaElement) {
										children.add((XmlSchemaElement) member);
									}
								}
								final ServiceObjectMap container = new ServiceObjectMap();
								for (int i = 0; i < children.size() && i < response.length; i++) {
									final XmlSchemaElement element = children.get(i);
									final String memberName = element.getName();
									final Object memberValue = ServiceObjectProxy.createProxyIfNecessary(response[i], element.getSchemaType());
									container.setProperty(memberName, memberValue);
								}
								responseObject.setProperty(partName, container);
							}
						} catch (final Exception e) {
							Logging.logWarning("Response object creation failed for operation '" + operationName + "'!", e, LOGGER);
						}
					}
				}
				return responseObject;
			}
		} catch (final Exception e) {
			Logging.logWarning("Service call on operation '" + operationName + "' failed!", e, LOGGER);
		}
		return null;
	}


	/**
	 * Creates a new object providing access to the structure of the specified message part.
	 *
	 * @param partInfo The message part to create.
	 * @return The new message part.
	 */
	@NotNull
	private ServiceObjectProxy createMessagePart(@NotNull final MessagePartInfo partInfo) {
		final String partName = partInfo.getName().getLocalPart();
		final String messageName = partInfo.getMessageInfo().getName().getLocalPart();
		final XmlSchemaAnnotated partSchema = partInfo.getXmlSchema();
		XmlSchemaType partSchemaType = null;
		if (partSchema instanceof XmlSchemaType) {
			partSchemaType = (XmlSchemaType) partSchema;
		} else if (partSchema instanceof XmlSchemaElement) {
			partSchemaType = ((XmlSchemaElement) partSchema).getSchemaType();
		}
		if (partSchemaType instanceof XmlSchemaComplexType) {
			final Class<?> typeClass = partInfo.getTypeClass();
			if (typeClass != null) {
				try {
					final Object object = typeClass.newInstance();
					prepareComplexType(object, partSchemaType);
					return new ServiceObjectProxy(object, (XmlSchemaComplexType) partSchemaType);
				} catch (final Exception e) {
					throw new IllegalStateException("Failed to create part '" + partName + "' of message '" + messageName + "'!", e);
				}
			}
			throw new IllegalStateException("Unknown type for part '" + partName + "' of message '" + messageName + "'!");
		}
		throw new IllegalStateException("Cannot create automatically simple part '" + partName + "' of message '" + messageName + "'!");
	}


	/**
	 * Prepares the specified object for use in this operation.
	 * This method will recursivly create the inner structure of the specified object,
	 * if it is based on a complex schema type definition.
	 *
	 * @param object The object to prepare for sending with this operation.
	 * @param schemaType The schema type the object is made of.
	 * @throws IntrospectionException
	 * @throws InstantiationException
	 * @throws IllegalAccessException
	 * @throws InvocationTargetException
	 */
	private static void prepareComplexType(@NotNull final Object object, @Nullable final XmlSchemaType schemaType) throws IntrospectionException, InstantiationException, IllegalAccessException, InvocationTargetException {
		if (schemaType instanceof XmlSchemaComplexType) {
			final XmlSchemaParticle particle = ((XmlSchemaComplexType) schemaType).getParticle();
			if (particle instanceof XmlSchemaSequence) {
				// Iterate through all members of the specified object, and create them if necessary
				for (final XmlSchemaSequenceMember member : ((XmlSchemaSequence) particle).getItems()) {
					if (member instanceof XmlSchemaElement) {
						if (((XmlSchemaElement) member).getMinOccurs() > 0) {
							// The member is mandatory
							//  => create the related inner structure if necessary
							final String memberName = ((XmlSchemaElement) member).getName();
							final PropertyDescriptor memberDescriptor = new PropertyDescriptor(memberName, object.getClass());
							final Class<?> memberType = memberDescriptor.getPropertyType();
							if (((XmlSchemaElement) member).getMaxOccurs() > 1) {
								// The member seems to be of type 'List' or 'Array'
								//  => Determine the actual list entry type
								final Type type = memberDescriptor.getReadMethod().getGenericReturnType();
								Type entryType = null;
								if (Collection.class.isAssignableFrom(memberType)) {
									if (type instanceof ParameterizedType) {
										entryType = ((ParameterizedType) type).getActualTypeArguments()[0];
									}
								} else if (memberType.isArray()) {
									if (type instanceof GenericArrayType) {
										entryType = ((GenericArrayType) type).getGenericComponentType();
									}
								}
								if (entryType instanceof Class<?>) {
									final List<Object> entries = new ArrayList<>();
									for (int i = 0; i < ((XmlSchemaElement) member).getMinOccurs(); i++) {
										final Object entryValue = ((Class<?>) entryType).newInstance();
										prepareComplexType(entryValue, ((XmlSchemaElement) member).getSchemaType());
										entries.add(entryValue);
									}
									if (Collection.class.isAssignableFrom(memberType)) {
										memberDescriptor.getWriteMethod().invoke(object, entries);
									} else if (memberType.isArray()) {
										memberDescriptor.getWriteMethod().invoke(object, entries.toArray());
									}
								}
							} else if (!memberType.isPrimitive() && !memberType.isEnum()) {
								// The member might be a complex type we need to initialized recursively
								final Object memberValue = memberType.newInstance();
								prepareComplexType(memberValue, ((XmlSchemaElement) member).getSchemaType());
								memberDescriptor.getWriteMethod().invoke(object, memberValue);
							}
						}
					}
				}
			}
		}
	}
}
